#-----------------------------
# -- kongodb --
#-----------------------------

from . import adapters, dictquery, lib, database, cursor, exceptions, types

# alias to Database
kongodb = database.Database


